from django.contrib import admin

from cardroom.models import CashGame, HandHistory, Poker, Table

admin.site.register(CashGame)
admin.site.register(HandHistory)
admin.site.register(Poker)
admin.site.register(Table)



from dataclasses import asdict
from typing import Any

from django.http import JsonResponse
from django.views.generic import DetailView
from rest_framework.viewsets import ModelViewSet

from cardroom.models import CashGame, HandHistory, Poker, Table
from cardroom.serializers import (
    CashGameSerializer,
    HandHistorySerializer,
    PokerSerializer,
    TableSerializer,
)
from cardroom.utilities import get_configuration, serialize


class PokerViewSet(ModelViewSet):
    queryset = Poker.objects.all()
    serializer_class = PokerSerializer


class TableViewSet(ModelViewSet):
    queryset = Table.objects.all()
    serializer_class = TableSerializer


class CashGameViewSet(ModelViewSet):
    queryset = CashGame.objects.all()
    serializer_class = CashGameSerializer


class HandHistoryViewSet(ModelViewSet):
    queryset = HandHistory.objects.all()
    serializer_class = HandHistorySerializer


class CashGameFeltView(DetailView):
    model = CashGame
    template_name = 'cardroom/cash-game.html'

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context['configuration'] = asdict(get_configuration())
        context['data'] = asdict(self.object.get_data(self.request.user))

        return context


class HandHistoryFeltView(DetailView):
    model = HandHistory
    template_name = 'cardroom/hand-history.html'

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context['configuration'] = asdict(get_configuration())
        context['data'] = serialize(self.object.data)

        return context


class CashGameFeltDataView(DetailView):
    model = CashGame

    def render_to_response(
            self,
            context: dict[str, Any],
            **response_kwargs: Any,
    ) -> JsonResponse:
        return JsonResponse(asdict(self.object.get_data(self.request.user)))


class HandHistoryFeltDataView(DetailView):
    model = HandHistory

    def render_to_response(
            self,
            context: dict[str, Any],
            **response_kwargs: Any,
    ) -> JsonResponse:
        return JsonResponse(serialize(self.object.data), safe=False)



from __future__ import annotations

from collections.abc import Iterator
from collections import deque
from dataclasses import dataclass, field
from random import choice
from warnings import warn

from pokerkit import Poker, State


@dataclass
class Button:
    table: Table
    seat_index: int | None = field(default=None, init=False)
    user: str | None = field(default=None, init=False)

    @property
    def seat(self) -> Seat | None:
        if self.seat_index is None:
            seat = None
        else:
            seat = self.table.seats[self.seat_index]

        return seat

    def step(self) -> deque[int]:
        seat_indices = deque[int]()

        if self.table.game.button_status:
            if (
                    self.user is None
                    or self.table.get_seat_or_none(self.user) is not self.seat
            ):
                self.seat_index = None

            if self.seat_index is None:
                self.seat_index = choice(
                    tuple(self.table.ready_or_postable_indices),
                )

                for seat in self.table.seats:
                    seat.wait_status = False
            else:
                seats = deque(self.table.seats)

                seats.rotate(-self.seat_index - 1)

                while not seats[0].ready_status:
                    seats[0].wait_status = False

                    seats.rotate(-1)

                self.seat_index = seats[0].index

            assert self.seat is not None

            self.user = self.seat.user
            seats = deque(self.table.seats)

            seats.rotate(-self.seat_index - 1)

            for seat in seats:
                if seat.ready_status:
                    seat_indices.append(seat.index)
        else:
            self.seat_index = None
            self.user = None

            for seat in self.table.seats:
                seat.wait_status = False

            seat_indices.extend(self.table.ready_indices)

        return seat_indices


@dataclass
class Seat:
    table: Table
    user: str | None = field(default=None, init=False)
    player_index: int | None = field(default=None, init=False)
    active_status: bool = field(default=False, init=False)
    wait_status: bool = field(default=False, init=False)
    starting_stack: int | None = field(default=None, init=False)

    @property
    def index(self) -> int:
        return next(
            i for i, seat in enumerate(self.table.seats) if self is seat
        )

    @property
    def user_status(self) -> bool:
        return self.user is not None

    @property
    def ready_or_postable_status(self) -> bool:
        return (
            self.user_status
            and self.active_status
            and self.starting_stack is not None
            and self.starting_stack > 0
        )

    @property
    def ready_status(self) -> bool:
        return self.ready_or_postable_status and not self.wait_status

    @property
    def player_status(self) -> bool:
        return self.player_index is not None


@dataclass
class Table:
    game: Poker
    seat_count: int
    min_starting_stack: int
    max_starting_stack: int
    button: Button = field(init=False)
    seats: list[Seat] = field(default_factory=list, init=False)
    state: State | None = field(default=None, init=False)

    def __post_init__(self) -> None:
        self.button = Button(self)

        for _ in range(self.seat_count):
            self.seats.append(Seat(self))

    @property
    def seat_indices(self) -> range:
        return range(self.seat_count)

    @property
    def user_indices(self) -> Iterator[int]:
        for seat in self.seats:
            if seat.user_status:
                yield seat.index

    @property
    def user_seats(self) -> Iterator[Seat]:
        for i in self.user_indices:
            yield self.seats[i]

    @property
    def user_count(self) -> int:
        return len(tuple(self.user_indices))

    @property
    def ready_or_postable_indices(self) -> Iterator[int]:
        for seat in self.seats:
            if seat.ready_or_postable_status:
                yield seat.index

    @property
    def ready_or_postable_seats(self) -> Iterator[Seat]:
        for i in self.ready_or_postable_indices:
            yield self.seats[i]

    @property
    def ready_or_postable_count(self) -> int:
        return len(tuple(self.ready_or_postable_indices))

    @property
    def ready_indices(self) -> Iterator[int]:
        for seat in self.seats:
            if seat.ready_status:
                yield seat.index

    @property
    def ready_seats(self) -> Iterator[Seat]:
        for i in self.ready_indices:
            yield self.seats[i]

    @property
    def ready_count(self) -> int:
        return len(tuple(self.ready_indices))

    @property
    def users(self) -> Iterator[str]:
        for seat in self.seats:
            if seat.user_status:
                assert seat.user is not None

                yield seat.user

    @property
    def acting_seat(self) -> Seat | None:
        if self.state is not None:
            for seat in self.seats:
                if (
                        seat.player_status
                        and (
                            (
                                seat.player_index
                                == self.state.stander_pat_or_discarder_index
                            )
                            or seat.player_index == self.state.actor_index
                            or seat.player_index == self.state.showdown_index
                        )
                ):
                    return seat

        return None

    def get_seat(self, user: str) -> Seat:
        for seat in self.seats:
            if seat.user == user:
                return seat

        raise ValueError('user not found')

    def get_seat_or_none(self, user: str) -> Seat | None:
        try:
            seat = self.get_seat(user)
        except ValueError:
            seat = None

        return seat

    # game changing

    def verify_game_changing(self, game: Poker) -> None:
        if self.state is not None:
            raise ValueError('ongoing game')

    def can_change_game(self, game: Poker) -> bool:
        try:
            self.verify_game_changing(game)
        except ValueError:
            return False

        return True

    def change_game(self, game: Poker) -> None:
        self.verify_game_changing(game)

        self.game = game

    # state management

    def verify_state_construction(self) -> None:
        if self.state is not None and self.state.status:
            raise ValueError('state already constructed')
        elif self.ready_or_postable_count < 2:
            raise ValueError('not enough ready players')

    def can_construct_state(self) -> bool:
        try:
            self.verify_state_construction()
        except ValueError:
            return False

        return True

    def construct_state(self) -> None:
        self.verify_state_construction()

        seat_indices = self.button.step()
        starting_stacks = []

        for i, j in enumerate(seat_indices):
            self.seats[j].player_index = i
            starting_stack = self.seats[j].starting_stack

            assert starting_stack is not None

            starting_stacks.append(starting_stack)

            self.seats[j].starting_stack = None

        player_count = len(seat_indices)
        self.state = self.game(starting_stacks, player_count)

    def verify_state_destruction(self) -> None:
        if self.state is None:
            raise ValueError('state already destroyed')
        elif self.state.status:
            raise ValueError('state active')

    def can_destroy_state(self) -> bool:
        try:
            self.verify_state_destruction()
        except ValueError:
            return False

        return True

    def destroy_state(self) -> None:
        self.verify_state_destruction()

        assert self.state
        assert not self.state.status

        for seat in self.seats:
            if seat.player_status:
                assert seat.player_index is not None

                stack = self.state.stacks[seat.player_index]

                if seat.starting_stack is None:
                    seat.starting_stack = stack
                elif stack > seat.starting_stack:
                    warn('rat-holing')

            seat.player_index = None

            if not seat.starting_stack:
                seat.starting_stack = None

        self.state = None

    # player management

    def verify_joining(self, user: str, seat_index: int | None = None) -> None:
        for seat in self.seats:
            if seat.user == user:
                raise ValueError('user already joined')

        if seat_index is None:
            if len(tuple(self.users)) == self.seat_count:
                raise ValueError('full table')
        else:
            if seat_index not in self.seat_indices:
                raise ValueError('invalid seat index')
            elif self.seats[seat_index].user_status:
                raise ValueError('seat occupied')

    def can_join(self, user: str, seat_index: int | None = None) -> bool:
        try:
            self.verify_joining(user, seat_index)
        except ValueError:
            return False

        return True

    def join(self, user: str, seat_index: int) -> None:
        self.verify_joining(user, seat_index)

        seat = self.seats[seat_index]
        seat.user = user
        seat.player_index = None
        seat.active_status = True
        seat.wait_status = True
        seat.starting_stack = None

    def verify_leaving(self, user: str) -> Seat:
        seat = self.get_seat(user)

        if seat.player_status:
            raise ValueError('player in hand')

        return seat

    def can_leave(self, user: str) -> bool:
        try:
            self.verify_leaving(user)
        except ValueError:
            return False

        return True

    def leave(self, user: str) -> None:
        seat = self.verify_leaving(user)

        assert not seat.player_status

        seat.user = None
        seat.player_index = None
        seat.active_status = False
        seat.wait_status = False
        seat.starting_stack = None

    def verify_sitting_out(self, user: str) -> Seat:
        seat = self.get_seat(user)

        if not seat.active_status:
            raise ValueError('already deactivated')

        return seat

    def can_sit_out(self, user: str) -> bool:
        try:
            self.verify_sitting_out(user)
        except ValueError:
            return False

        return True

    def sit_out(self, user: str) -> None:
        seat = self.verify_sitting_out(user)
        seat.active_status = False

    def verify_being_back(self, user: str) -> Seat:
        seat = self.get_seat(user)

        if seat.active_status:
            raise ValueError('already activated')

        return seat

    def can_be_back(self, user: str) -> bool:
        try:
            self.verify_being_back(user)
        except ValueError:
            return False

        return True

    def be_back(self, user: str) -> None:
        seat = self.verify_being_back(user)
        seat.active_status = True

    def verify_buying_rebuying_topping_off_or_rat_holing(
            self,
            user: str,
            starting_stack: int | None = None,
    ) -> Seat:
        seat = self.get_seat(user)

        if not seat.active_status:
            raise ValueError('inactive user')

        if starting_stack is not None:
            if starting_stack < self.min_starting_stack:
                raise ValueError('below minimum starting stack')
            elif starting_stack > self.max_starting_stack:
                raise ValueError('above maximum starting stack')
            elif (
                    seat.starting_stack is not None
                    and seat.starting_stack > starting_stack
            ):
                warn('rat-holing')

        return seat

    def can_buy_rebuy_top_off_or_rat_hole(
            self,
            user: str,
            starting_stack: int | None = None,
    ) -> bool:
        try:
            self.verify_buying_rebuying_topping_off_or_rat_holing(
                user,
                starting_stack,
            )
        except ValueError:
            return False

        return True

    def buy_rebuy_top_off_or_rat_hole(
            self,
            user: str,
            starting_stack: int,
    ) -> None:
        seat = self.verify_buying_rebuying_topping_off_or_rat_holing(
            user,
            starting_stack,
        )
        seat.starting_stack = starting_stack



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from queue import Empty, Queue
from traceback import print_exc
from typing import Any
from warnings import warn
from zoneinfo import ZoneInfo

from pokerkit import min_or_none, parse_action

from cardroom.felt import Data
from cardroom.table import Table


@dataclass
class Controller(ABC):
    time_bank: float
    """The time bank."""
    time_bank_increment: float
    """The time bank increment."""
    state_construction_timeout: float
    """The state construction timeout."""
    state_destruction_timeout: float
    """The state destruction timeout."""
    idle_timeout: float
    """The idle timeout."""
    standing_pat_timeout: float
    """The standing pat timeout."""
    betting_timeout: float
    """The betting timeout."""
    hole_cards_showing_or_mucking_timeout: float
    """The hole cards showing or mucking timeout."""
    callback: Callable[[list[dict[str, Data]], tuple[str, str] | None], Any]
    """The callback."""
    parse_value: Callable[[str], int]
    """The value parser."""
    tzinfo: ZoneInfo
    """The timezone."""

    @abstractmethod
    def mainloop(self) -> None:
        pass

    @abstractmethod
    def handle(self, user: str, event: Any) -> None:
        pass

    def _run(self, table: Table, queue: Queue[tuple[str, Any]]) -> None:

        def get_now_timestamp() -> datetime:
            return datetime.now(self.tzinfo)

        def is_past_timestamp(timestamp: datetime | None) -> bool:
            if timestamp is None:
                status = False
            else:
                status = get_now_timestamp() >= timestamp

            return status

        def get_future_timestamp(timeout: float) -> datetime:
            return get_now_timestamp() + timedelta(seconds=timeout)

        def get_auto_timestamp() -> datetime | None:
            return min_or_none(
                (
                    state_construction_timestamp,
                    state_destruction_timestamp,
                    *idle_timestamps.values(),
                    standing_pat_timestamp,
                    betting_timestamp,
                    hole_cards_showing_or_mucking_timestamp,
                ),
            )

        def get_event() -> tuple[str, Any] | None:
            if (auto_timestamp := get_auto_timestamp()) is not None:
                timeout = max(
                    (auto_timestamp - get_now_timestamp()).total_seconds(),
                    0,
                )
            else:
                timeout = None

            try:
                event = queue.get(timeout=timeout)
            except Empty:
                event = None

            return event

        def append_data() -> None:
            data.append(Data.from_table(table))

        def get_time_bank(user: str) -> float:
            time_banks.setdefault(user, self.time_bank)

            return time_banks[user]

        state_construction_timestamp = None
        state_destruction_timestamp = None
        idle_timestamps = dict[str, datetime | None]()
        standing_pat_timestamp: datetime | None = None
        betting_timestamp: datetime | None = None
        hole_cards_showing_or_mucking_timestamp: datetime | None = None

        def parse_user_action() -> None:
            nonlocal standing_pat_timestamp
            nonlocal betting_timestamp
            nonlocal hole_cards_showing_or_mucking_timestamp

            tokens = action.split()

            match tokens:
                case 'j', seat_index:
                    table.join(user, int(seat_index))
                case ('l',):
                    table.leave(user)
                case ('s',):
                    table.sit_out(user)
                case ('b',):
                    table.be_back(user)
                case 'brtr', starting_stack:
                    table.buy_rebuy_top_off_or_rat_hole(
                        user,
                        self.parse_value(starting_stack),
                    )
                case _:
                    player_index = table.get_seat(user).player_index

                    if player_index is None:
                        raise ValueError('player dne')

                    player = f'p{player_index + 1}'

                    if table.state is None:
                        raise ValueError('state dne')

                    if 'sm' in (tokens := action.split()):
                        if '#' in tokens:
                            tokens = tokens[:tokens.index('#')]

                        if tokens != ['sm'] or tokens != ['sm', '-']:
                            raise ValueError('explicit showdown is forbidden')

                    parse_action(
                        table.state,
                        f'{player} {action}',
                        self.parse_value,
                    )

                    standing_pat_timestamp = None
                    betting_timestamp = None
                    hole_cards_showing_or_mucking_timestamp = None

        time_banks = dict[str, float]()
        data = list[dict[str, Data]]()
        user_error = None

        append_data()
        self.callback(data, user_error)
        data.clear()

        while True:
            user_action = get_event()

            if user_action is not None:
                user, action = user_action

                if isinstance(action, str):
                    try:
                        parse_user_action()
                    except ValueError as exception:
                        user_error = user, str(exception)

                        print_exc()
                    else:
                        append_data()
                else:
                    warn('cannot handle event')

            data_count = None

            while data_count != len(data):
                data_count = len(data)

                for user in table.users:
                    seat = table.get_seat(user)

                    if (
                            table.state is not None
                            and not seat.player_status
                            and table.can_sit_out(user)
                    ):
                        table.sit_out(user)

                    status = is_past_timestamp(idle_timestamps.get(user))

                    if (
                            status
                            or seat.active_status
                            or not table.can_leave(user)
                    ):
                        idle_timestamps[user] = None
                    elif idle_timestamps.get(user) is None:
                        idle_timestamps[user] = get_future_timestamp(
                            self.idle_timeout,
                        )

                    if (
                            status
                            and not seat.active_status
                            and table.can_leave(user)
                    ):
                        table.leave(user)
                        append_data()

                status = is_past_timestamp(state_construction_timestamp)

                if status or not table.can_construct_state():
                    state_construction_timestamp = None
                elif state_construction_timestamp is None:
                    state_construction_timestamp = get_future_timestamp(
                        self.state_construction_timeout,
                    )

                if status and table.can_construct_state():
                    table.construct_state()
                    append_data()

                status = is_past_timestamp(state_destruction_timestamp)

                if status or not table.can_destroy_state():
                    state_destruction_timestamp = None
                elif state_destruction_timestamp is None:
                    state_destruction_timestamp = get_future_timestamp(
                        self.state_destruction_timeout,
                    )

                if status and table.can_destroy_state():
                    for key, value in time_banks.items():
                        time_banks[key] = min(
                            self.time_bank,
                            value + self.time_bank_increment,
                        )

                    table.destroy_state()
                    append_data()

                if table.state is None:
                    standing_pat_timestamp = None
                    betting_timestamp = None
                    hole_cards_showing_or_mucking_timestamp = None
                else:
                    if (
                            table.acting_seat is not None
                            and not table.acting_seat.active_status
                    ):
                        if table.state.can_stand_pat_or_discard():
                            table.state.stand_pat_or_discard()
                        elif table.state.can_post_bring_in():
                            table.state.post_bring_in()
                        elif table.state.can_fold():
                            table.state.fold()
                        elif table.state.can_check_or_call():
                            table.state.check_or_call()
                        elif table.state.can_show_or_muck_hole_cards():
                            table.state.show_or_muck_hole_cards()
                        else:
                            raise AssertionError

                        append_data()

                    status = True

                    if table.state.can_post_ante():
                        table.state.post_ante()
                    elif table.state.can_collect_bets():
                        table.state.collect_bets()
                    elif table.state.can_post_blind_or_straddle():
                        table.state.post_blind_or_straddle()
                    elif table.state.can_deal_board():
                        table.state.deal_board()
                    elif table.state.can_deal_hole():
                        table.state.deal_hole()
                    elif table.state.can_kill_hand():
                        table.state.kill_hand()
                    elif table.state.can_push_chips():
                        table.state.push_chips()
                    elif table.state.can_pull_chips():
                        table.state.pull_chips()
                    else:
                        status = False

                    if status:
                        append_data()

                    status = is_past_timestamp(standing_pat_timestamp)

                    if status or not table.state.can_stand_pat_or_discard():
                        standing_pat_timestamp = None
                    else:
                        standing_pat_timestamp = get_future_timestamp(
                            self.standing_pat_timeout,
                        )

                    if status and table.state.can_stand_pat_or_discard():
                        table.state.stand_pat_or_discard()
                        append_data()

                    status = is_past_timestamp(betting_timestamp)

                    if status or table.state.actor_index is None:
                        betting_timestamp = None
                    else:
                        betting_timestamp = get_future_timestamp(
                            self.betting_timeout,
                        )

                    if status and table.state.actor_index is not None:
                        if table.state.can_fold():
                            table.state.fold()
                        elif table.state.can_check_or_call():
                            table.state.check_or_call()
                        elif table.state.can_post_bring_in():
                            table.state.post_bring_in()
                        else:
                            raise AssertionError

                        append_data()

                    status = is_past_timestamp(
                        hole_cards_showing_or_mucking_timestamp,
                    )

                    if status or not table.state.can_show_or_muck_hole_cards():
                        hole_cards_showing_or_mucking_timestamp = None
                    elif hole_cards_showing_or_mucking_timestamp is None:
                        hole_cards_showing_or_mucking_timestamp = (
                            get_future_timestamp(
                                self.hole_cards_showing_or_mucking_timeout,
                            )
                        )

                    if status and table.state.can_show_or_muck_hole_cards():
                        table.state.show_or_muck_hole_cards()
                        append_data()

            for user in set(idle_timestamps) - set(table.users):
                idle_timestamps.pop(user)

            for user in set(time_banks) - set(table.users):
                time_banks.pop(user)

            # TODO: send over
            # timestamp = get_now_timestamp()
            # auto_timestamp = get_auto_timestamp()

            if data or user_error is not None:
                self.callback(data, user_error)

            data.clear()
            user_error = None


@dataclass
class CashGame(Controller):
    table: Table
    """The table."""
    queue: Queue[tuple[str, Any]] = field(default_factory=Queue, init=False)
    """The queue."""

    def mainloop(self) -> None:
        self._run(self.table, self.queue)

    def handle(self, user: str, event: Any) -> None:
        self.queue.put((user, event))


@dataclass
class Tournament(Controller):
    def mainloop(self) -> None:
        raise NotImplementedError

    def handle(self, user: str, event: Any) -> None:
        raise NotImplementedError



from importlib import import_module

from channels.auth import AuthMiddlewareStack
from channels.routing import ProtocolTypeRouter, URLRouter
from django.core.asgi import get_asgi_application

from cardroom.utilities import get_root_routingconf

application = ProtocolTypeRouter(
    {
        'http': get_asgi_application(),
        'websocket': AuthMiddlewareStack(
            URLRouter(import_module(get_root_routingconf()).urlpatterns),
        )
    },
)



from warnings import resetwarnings, simplefilter

from django.test import SimpleTestCase
from pokerkit import FixedLimitRazz, NoLimitTexasHoldem, State

from cardroom.table import Table


class TableTestCase(SimpleTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        simplefilter('ignore')

    @classmethod
    def tearDownClass(cls) -> None:
        resetwarnings()

    def create_nt_table(self) -> Table:
        return Table(
            NoLimitTexasHoldem((), False, {1: 3}, [1, 2], 2),
            6,
            80,
            200,
        )

    def create_fr_table(self) -> Table:
        return Table(
            FixedLimitRazz((), True, 1, 1, 4, 8),
            6,
            80,
            200,
        )

    def terminate(self, state: State) -> None:
        while state.status:
            if state.can_post_ante():
                state.post_ante()
            elif state.can_post_blind_or_straddle():
                state.post_blind_or_straddle()
            elif state.can_collect_bets():
                state.collect_bets()
            elif state.can_burn_card():
                state.burn_card()  # pragma: no cover
            elif state.can_deal_board():
                state.deal_board()  # pragma: no cover
            elif state.can_deal_hole():
                state.deal_hole()
            elif state.can_stand_pat_or_discard():
                state.stand_pat_or_discard()  # pragma: no cover
            elif state.can_fold():
                state.fold()
            elif state.can_check_or_call():
                state.check_or_call()  # pragma: no cover
            elif state.can_post_bring_in():
                state.post_bring_in()
            elif state.can_complete_bet_or_raise_to():
                state.complete_bet_or_raise_to()  # pragma: no cover
            elif state.can_show_or_muck_hole_cards():
                state.show_or_muck_hole_cards()  # pragma: no cover
            elif state.can_kill_hand():
                state.kill_hand()  # pragma: no cover
            elif state.can_push_chips():
                state.push_chips()
            elif state.can_pull_chips():
                state.pull_chips()

    def test_game_changing(self) -> None:
        table = self.create_nt_table()
        game0 = table.game
        game1 = NoLimitTexasHoldem((), True, 0, [1, 2], 2)

        self.assertTrue(table.can_change_game(game0))
        self.assertTrue(table.can_change_game(game1))
        table.change_game(game1)
        self.assertTrue(table.can_change_game(game0))
        self.assertTrue(table.can_change_game(game1))
        self.assertIsNot(table.game, game0)
        self.assertIs(table.game, game1)

    def test_state_construction_and_state_deconstruction(self) -> None:
        table = self.create_nt_table()

        self.assertFalse(table.can_construct_state())
        table.join('u0', 4)
        self.assertFalse(table.can_construct_state())
        table.join('u1', 5)
        self.assertFalse(table.can_construct_state())
        table.buy_rebuy_top_off_or_rat_hole('u0', 200)
        self.assertFalse(table.can_construct_state())
        table.buy_rebuy_top_off_or_rat_hole('u1', 200)
        self.assertTrue(table.can_construct_state())
        table.sit_out('u0')
        self.assertFalse(table.can_construct_state())
        table.be_back('u0')
        self.assertTrue(table.can_construct_state())
        table.join('u2', 3)
        self.assertTrue(table.can_construct_state())
        table.join('u3', 0)
        self.assertTrue(table.can_construct_state())
        self.assertFalse(table.can_destroy_state())

        u0_indices = set()
        u2_indices = set()

        table.construct_state()

        assert table.state is not None

        u0_indices.add(table.get_seat('u0').player_index)
        u2_indices.add(table.get_seat('u2').player_index)
        self.assertEqual(table.state.starting_stacks, (200,) * 2)
        self.assertEqual(table.state.player_count, 2)
        self.assertFalse(table.can_destroy_state())
        self.terminate(table.state)
        self.assertTrue(table.can_destroy_state())
        table.destroy_state()

        table.construct_state()

        assert table.state is not None

        u0_indices.add(table.get_seat('u0').player_index)
        u2_indices.add(table.get_seat('u2').player_index)
        self.assertNotEqual(table.state.starting_stacks, (200,) * 2)
        self.assertEqual(table.state.player_count, 2)
        table.buy_rebuy_top_off_or_rat_hole('u2', 200)
        self.terminate(table.state)
        table.buy_rebuy_top_off_or_rat_hole('u3', 200)
        table.destroy_state()

        table.construct_state()

        assert table.state is not None

        u0_indices.add(table.get_seat('u0').player_index)
        u2_indices.add(table.get_seat('u2').player_index)
        self.terminate(table.state)
        table.destroy_state()

        table.construct_state()

        assert table.state is not None

        u0_indices.add(table.get_seat('u0').player_index)
        u2_indices.add(table.get_seat('u2').player_index)
        self.assertEqual(table.state.player_count, 4)
        self.terminate(table.state)
        table.buy_rebuy_top_off_or_rat_hole('u0', 200)
        table.buy_rebuy_top_off_or_rat_hole('u3', 200)
        table.destroy_state()

        table.buy_rebuy_top_off_or_rat_hole('u1', 200)
        table.buy_rebuy_top_off_or_rat_hole('u2', 100)
        table.buy_rebuy_top_off_or_rat_hole('u2', 200)

        table.construct_state()

        assert table.state is not None

        table.buy_rebuy_top_off_or_rat_hole('u0', 100)
        u0_indices.add(table.get_seat('u0').player_index)
        u2_indices.add(table.get_seat('u2').player_index)
        self.assertEqual(table.state.starting_stacks, (200,) * 4)
        self.terminate(table.state)
        table.destroy_state()

        self.assertNotIn(None, u0_indices)
        self.assertIn(None, u2_indices)
        self.assertGreaterEqual(len(u0_indices), 2)
        self.assertGreaterEqual(len(u0_indices), 3)

        table = self.create_fr_table()

        table.join('u0', 0)
        table.join('u1', 1)
        table.join('u2', 2)
        table.join('u3', 3)
        table.buy_rebuy_top_off_or_rat_hole('u0', 200)
        table.buy_rebuy_top_off_or_rat_hole('u1', 200)

        table.construct_state()

        assert table.state is not None

        self.assertEqual(table.get_seat('u0').player_index, 0)
        self.assertEqual(table.get_seat('u1').player_index, 1)
        self.assertEqual(table.get_seat('u2').player_index, None)
        self.assertEqual(table.get_seat('u3').player_index, None)
        self.terminate(table.state)
        table.destroy_state()

        table.construct_state()

        assert table.state is not None

        self.assertEqual(table.get_seat('u0').player_index, 0)
        self.assertEqual(table.get_seat('u1').player_index, 1)
        self.assertEqual(table.get_seat('u2').player_index, None)
        self.assertEqual(table.get_seat('u3').player_index, None)
        table.buy_rebuy_top_off_or_rat_hole('u2', 200)
        self.terminate(table.state)
        table.buy_rebuy_top_off_or_rat_hole('u3', 200)
        table.destroy_state()

        table.construct_state()

        assert table.state is not None

        self.assertEqual(table.get_seat('u0').player_index, 0)
        self.assertEqual(table.get_seat('u1').player_index, 1)
        self.assertEqual(table.get_seat('u2').player_index, 2)
        self.assertEqual(table.get_seat('u3').player_index, 3)
        self.terminate(table.state)
        table.destroy_state()

    def test_sitting_and_leaving(self) -> None:
        table = self.create_nt_table()

        self.assertTrue(table.can_join('u0', 4))
        table.join('u0', 4)
        self.assertFalse(table.can_join('u0', 4))
        self.assertFalse(table.can_join('u0', 5))
        self.assertFalse(table.can_join('u1', 4))
        self.assertTrue(table.can_join('u1', 5))
        table.join('u1', 5)
        self.assertFalse(table.can_join('u2', 4))
        self.assertTrue(table.can_leave('u0'))
        table.leave('u0')
        self.assertTrue(table.can_join('u2', 4))
        table.join('u2', 4)
        self.assertFalse(table.can_leave('u0'))
        table.join('u0', 2)
        self.assertTrue(table.can_leave('u0'))
        table.leave('u0')

    def test_sitting_out_and_being_back(self) -> None:
        table = self.create_nt_table()

        self.assertFalse(table.can_sit_out('u0'))
        self.assertFalse(table.can_be_back('u0'))
        table.join('u0', 4)
        self.assertTrue(table.can_sit_out('u0'))
        self.assertFalse(table.can_be_back('u0'))
        table.sit_out('u0')
        self.assertFalse(table.can_sit_out('u0'))
        self.assertTrue(table.can_be_back('u0'))
        table.be_back('u0')
        self.assertTrue(table.can_sit_out('u0'))
        self.assertFalse(table.can_be_back('u0'))
        table.leave('u0')
        self.assertFalse(table.can_sit_out('u0'))
        self.assertFalse(table.can_be_back('u0'))

    def test_rebuying_or_topping_off(self) -> None:
        table = self.create_nt_table()

        self.assertFalse(table.can_buy_rebuy_top_off_or_rat_hole('u0', 200))
        table.join('u0', 5)
        self.assertFalse(table.can_buy_rebuy_top_off_or_rat_hole('u0', 40))
        self.assertTrue(table.can_buy_rebuy_top_off_or_rat_hole('u0', 100))
        self.assertTrue(table.can_buy_rebuy_top_off_or_rat_hole('u0', 150))
        self.assertTrue(table.can_buy_rebuy_top_off_or_rat_hole('u0', 200))
        self.assertFalse(table.can_buy_rebuy_top_off_or_rat_hole('u0', 400))
        table.buy_rebuy_top_off_or_rat_hole('u0', 150)
        self.assertFalse(table.can_buy_rebuy_top_off_or_rat_hole('u0', 40))
        self.assertTrue(table.can_buy_rebuy_top_off_or_rat_hole('u0', 100))
        self.assertTrue(table.can_buy_rebuy_top_off_or_rat_hole('u0', 150))
        self.assertTrue(table.can_buy_rebuy_top_off_or_rat_hole('u0', 200))
        self.assertFalse(table.can_buy_rebuy_top_off_or_rat_hole('u0', 400))
        table.buy_rebuy_top_off_or_rat_hole('u0', 200)
        self.assertFalse(table.can_buy_rebuy_top_off_or_rat_hole('u0', 40))
        self.assertTrue(table.can_buy_rebuy_top_off_or_rat_hole('u0', 100))
        self.assertTrue(table.can_buy_rebuy_top_off_or_rat_hole('u0', 150))
        self.assertTrue(table.can_buy_rebuy_top_off_or_rat_hole('u0', 200))
        self.assertFalse(table.can_buy_rebuy_top_off_or_rat_hole('u0', 400))



from math import inf

from django.conf import settings
from django.test import SimpleTestCase
from django.test.utils import override_settings

_divmod = divmod

from cardroom.utilities import (  # noqa: E402
    divmod,
    get_admin,
    get_auth,
    get_decimal_places,
    get_divmod,
    get_felt,
    get_parse_value,
    parse_value,
)


class UtilitiesTestCase(SimpleTestCase):
    @override_settings(CARDROOM_DECIMAL_PLACES=0)
    def test_zero_decimal_places(self) -> None:
        del settings.CARDROOM_DIVMOD
        del settings.CARDROOM_PARSE_VALUE

        self.assertEqual(get_divmod()(3, 2), (1, 1))
        self.assertEqual(get_divmod()(8, 3), (2, 2))
        self.assertEqual(get_parse_value()('0'), 0)
        self.assertEqual(get_parse_value()('1.25'), 1)
        self.assertEqual(get_parse_value()(str(8 / 3)), 2)

    @override_settings(CARDROOM_DECIMAL_PLACES=2)
    def test_two_decimal_places(self) -> None:
        del settings.CARDROOM_DIVMOD
        del settings.CARDROOM_PARSE_VALUE

        self.assertEqual(get_divmod()(3, 2), (1.5, 0))
        self.assertEqual(get_divmod()(8, 3)[0], 2.66)
        self.assertAlmostEqual(get_divmod()(8, 3)[1], 0.02)
        self.assertEqual(
            get_divmod()(3.0, 2.0),  # type: ignore[arg-type]
            (1.5, 0.0),
        )
        self.assertEqual(
            get_divmod()(8.0, 3.0)[0],  # type: ignore[arg-type]
            2.66,
        )
        self.assertAlmostEqual(  # type: ignore[misc]
            get_divmod()(8.0, 3.0)[1],  # type: ignore[arg-type]
            0.02,
        )
        self.assertEqual(get_parse_value()('0'), 0.0)
        self.assertEqual(get_parse_value()('1.25'), 1.25)
        self.assertEqual(get_parse_value()(str(8 / 3)), 2.67)

    @override_settings(CARDROOM_DECIMAL_PLACES=inf)
    def test_inf_decimal_places(self) -> None:
        del settings.CARDROOM_DIVMOD
        del settings.CARDROOM_PARSE_VALUE

        self.assertEqual(get_divmod()(3, 2), (1.5, 0))
        self.assertEqual(get_divmod()(8, 3), (8 / 3, 0))
        self.assertEqual(
            get_divmod()(3.0, 2.0),  # type: ignore[arg-type]
            (1.5, 0.0),
        )
        self.assertEqual(
            get_divmod()(8.0, 3.0),  # type: ignore[arg-type]
            (8 / 3, 0.0),
        )
        self.assertEqual(get_parse_value()('0'), 0)
        self.assertEqual(get_parse_value()('1.25'), 1.25)
        self.assertEqual(get_parse_value()(str(8 / 3)), 8 / 3)

    @override_settings(
            CARDROOM_DIVMOD='builtins.divmod',
            CARDROOM_PARSE_VALUE='builtins.int',
            CARDROOM_DECIMAL_PLACES=0,
            CARDROOM_ADMIN=True,
            CARDROOM_AUTH=True,
            CARDROOM_FELT=True,
    )
    def test_non_defaults_0(self) -> None:
        self.assertIs(get_divmod(), _divmod)
        self.assertIs(get_parse_value(), int)
        self.assertEqual(get_decimal_places(), 0)
        self.assertTrue(get_admin())
        self.assertTrue(get_auth())
        self.assertTrue(get_felt())

    @override_settings(
            CARDROOM_DIVMOD='cardroom.utilities.divmod',
            CARDROOM_PARSE_VALUE='cardroom.utilities.parse_value',
            CARDROOM_DECIMAL_PLACES=2,
            CARDROOM_ADMIN=False,
            CARDROOM_AUTH=False,
            CARDROOM_FELT=False,
    )
    def test_non_defaults_1(self) -> None:
        self.assertIs(get_divmod(), divmod)
        self.assertIs(get_parse_value(), parse_value)
        self.assertEqual(get_decimal_places(), 2)
        self.assertFalse(get_admin())
        self.assertFalse(get_auth())
        self.assertFalse(get_felt())

    @override_settings()
    def test_defaults(self) -> None:
        del settings.CARDROOM_DIVMOD
        del settings.CARDROOM_PARSE_VALUE
        del settings.CARDROOM_DECIMAL_PLACES
        del settings.CARDROOM_ADMIN
        del settings.CARDROOM_AUTH
        del settings.CARDROOM_FELT

        self.assertIs(get_divmod(), divmod)
        self.assertIs(get_parse_value(), parse_value)
        self.assertEqual(get_decimal_places(), inf)
        self.assertFalse(get_admin())
        self.assertFalse(get_auth())
        self.assertFalse(get_felt())



from django.apps import AppConfig


class CardroomConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'cardroom'

    def ready(self) -> None:
        from cardroom.models import CashGame
        from cardroom.gamemaster import set_controller

        __import__('cardroom.signals')

        for cash_game in CashGame.objects.all():
            set_controller(cash_game)



from threading import Thread, Lock

from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer

from cardroom.utilities import serialize

controller_lock = Lock()
controllers = {}
threads = {}


def set_controller(model):
    controllers[model.group_name] = model.load()
    threads[model.group_name] = Thread(
        target=controllers[model.group_name].mainloop,
        daemon=True,
    )

    threads[model.group_name].start()


def get_controller(model):
    with controller_lock:
        return controllers[model.group_name]


def handle(model, user, content):
    get_controller(model).handle(user.username, content)


data_lock = Lock()
data = {}


def broadcast(model, data_, user_message):
    if data_ is not None and data_:
        with data_lock:
            data[model.group_name] = data_[-1]

        async_to_sync(get_channel_layer().group_send)(
            model.group_name,
            {'type': 'data', 'data': serialize(data_)},
        )

    if user_message is not None:
        user, message = user_message

        async_to_sync(get_channel_layer().group_send)(
            model.group_name,
            {'type': 'message', 'user': user, 'message': message},
        )


def get_data(model):
    with data_lock:
        return data[model.group_name]
